/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Messages module.
 * 
 * This module allows to manipulate the msg_t and msg_avp_t structures that represents a Diameter message in memory.
 * See message.h and message-api.h for more information on the functions and types involved.
 */

#include <errno.h>
#include <string.h>
#include <assert.h>

#include "waaad-internal.h"

/* These types are internal to this module. It represents the "real" content of msg_t and msg_avp_t objects. */

/* Forward declaration */
struct _msg_avp_chain;

/* List structure */
typedef struct _msg_list {
	struct _msg_list      *next;  /* Next element in the list */
	struct _msg_list      *prev;  /* Previous element in the list */
	struct _msg_list      *head;  /* Pointer to the sentinel of this list */
	struct _msg_avp_chain *top;   /* the head of the current object */
	rule_position_t	       pos;   /* The position rule of this object, if any, otherwise 0. */
} _msg_list_t;

/* Return boolean, true if the list is empty */
#define IS_LIST_EMPTY( _list ) ( ((_list)->head == (_list)) && ((_list)->next == (_list)))

typedef enum {
	_MSG_MSG = 1, 
	_MSG_AVP
} _msg_objtype_t;

/* The following structure is used to represent the chaining of AVPs in a message or a grouped AVP. */
typedef struct _msg_avp_chain {
	_msg_list_t		chaining;	/* Chaining information at this level. */
	_msg_list_t		children;	/* sentinel for the children of this object */
	_msg_objtype_t		type;		/* Type of this object, _MSG_MSG or _MSG_AVP */
} _msg_avp_chain_t;


/* Return the chain information from an AVP or MSG. Since it's the first field, we just cast */
#define _C(_x) ((_msg_avp_chain_t *)(_x))


/* Some details about chaining:
 *
 *  A message is made of a header ( _msg_t ) and 0 or more AVPs (_msg_avp_t ).
 * The structure is a kind of tree, where AVPs can contain other AVPs (grouped AVPs).
 * Exemple:
 * msg
 *  |-avp
 *  |-gavp
 *  |   |-avp
 *  |   |-avp
 *  |   \-avp
 *  |-avp
 *  \-avp
 *
 * Each "component" (msg or avp) structure begins with a _msg_avp_chain_t component.
 * The element at the top of the hierarchy (msg in our example) has all the fields of its "chaining" equal to the same value.
 *
 * All elements at the same level are linked by their "chaining" list.
 * The "children" list is the sentinel for the lists of children of this element.
 */

/* The following definitions are used to avoid programming errors. */
#define _MSG_MSG_EYEC	(0x11355463)
#define _MSG_AVP_EYEC	(0x11355467)

/* The following structure represents an AVP instance. */
typedef struct {
	_msg_avp_chain_t avp_chain;		/* Chaining information of this AVP */
	int		 avp_eyec;		/* Must be equal to _MSG_AVP_EYEC */
	dict_object_t 	*avp_model;		/* If not NULL, pointer to the dictionary object of this avp */
	msg_avp_data_t	 avp_public;		/* AVP data that can be managed by extensions */
	
	char		*avp_source;		/* If the message was parsed from a buffer, pointer to the AVP start in the buffer. */
	avp_value_t	 avp_storage;		/* To avoid many alloc/free, store the integer values here and set avp_public.avp_data to &storage */
	int		 avp_mustfreeos;	/* 1 if an octetstring is malloc'd in avp_storage and must be freed. */
} _msg_avp_t;

/* Macro to compute the AVP header size */
#define AVPHDRSZ_NOVEND	8
#define AVPHDRSZ_VENDOR	12
#define GETAVPHDRSZ( _vend_flag ) ((_vend_flag) ? AVPHDRSZ_VENDOR : AVPHDRSZ_NOVEND)

/* Macro to cast a msg_avp_t */
#define _A(_x) ((_msg_avp_t *)(_x))
/* Check the type and eyecatcher */
#define CHECK_AVP(_x) ((_C(_x)->type == _MSG_AVP) && (_A(_x)->avp_eyec == _MSG_AVP_EYEC))

/* The following structure represents an instance of a message (command and children AVPs). */
typedef struct {
	_msg_avp_chain_t msg_chain;		/* List of the AVPs in the message */
	int		 msg_eyec;		/* Must be equal to _MSG_MSG_EYEC */
	dict_object_t 	*msg_model;		/* If not NULL, pointer to the dictionary object of this message */
	msg_data_t	 msg_public;		/* Message data that can be managed by extensions. */
	
	char		*msg_source;		/* If the message was parsed from a buffer, pointer to the message start in the buffer. */
	int		 msg_routable;		/* Is this a routable message? (0: undef, 1: routable, 2: non routable) */
} _msg_t;

/* Macro to compute the message header size */
#define GETMSGHDRSZ() 	20

/* Macro to cast a msg_avp_t */
#define _M(_x) ((_msg_t *)(_x))
/* Check the type and eyecatcher */
#define CHECK_MSG(_x) ((_C(_x)->type == _MSG_MSG) && (_M(_x)->msg_eyec == _MSG_MSG_EYEC))

#define VALIDATE_OBJ(_x) ( (CHECK_MSG(_x)) || (CHECK_AVP(_x)) )


/* Macro to validate a MSGFL_ value */
#define CHECK_MSGFL(_fl) ( ((_fl) & (- (MSGFL_MAX << 1) )) == 0 )

/* initial sizes of AVP from their types, in bytes. -1 for unknown sizes */
int avp_value_sizes[] = { 
	 0,	/* AVP_TYPE_GROUPED: size is dynamic */
	 0,	/* AVP_TYPE_OCTETSTRING: size is dynamic */
	 4,	/* AVP_TYPE_INTEGER32: size is 32 bits */
	 8,	/* AVP_TYPE_INTEGER64: size is 64 bits */
	 4,	/* AVP_TYPE_UNSIGNED32: size is 32 bits */
	 8,	/* AVP_TYPE_UNSIGNED64: size is 64 bits */
	 4,	/* AVP_TYPE_FLOAT32: size is 32 bits */
	 8	/* AVP_TYPE_FLOAT64: size is 64 bits */
};
#define CHECK_BASETYPE( _type ) ( (_type <= AVP_TYPE_MAX) && (_type >= 0) )
#define GETINITIALSIZE( _type, _vend ) (avp_value_sizes[ CHECK_BASETYPE(_type) ? _type : 0] + GETAVPHDRSZ(_vend))

/***************************************************************************************************************/

static void item_unlink(_msg_list_t * item);

/* Initialize a _msg_list_t structure */
static void init_list(_msg_list_t * list, _msg_avp_chain_t * top)
{
	list->next = list;
	list->prev = list;
	list->head = list;
	list->top  = top;
}

/* Initialize a _msg_avp_chain_t structure */
static void init_chain(_msg_avp_chain_t * chain, int type)
{
	init_list( &chain->chaining, chain);
	init_list( &chain->children, chain);
	chain->type = type;
}

/* Initialize a new AVP object */
static void init_avp ( _msg_avp_t * avp )
{
	memset(avp, 0, sizeof(_msg_avp_t));
	init_chain( &avp->avp_chain, _MSG_AVP);
	avp->avp_eyec = _MSG_AVP_EYEC;
}
	
/* Initialize a new MSG object */
static void init_msg ( _msg_t * msg )
{
	memset(msg, 0, sizeof(_msg_t));
	init_chain( &msg->msg_chain, _MSG_MSG);
	msg->msg_eyec = _MSG_MSG_EYEC;
}

/* Destroy and free an AVP or message */
static int destroy_obj (_msg_avp_chain_t * obj )
{
	/* Check the parameter is a valid object */
	if (!VALIDATE_OBJ(obj)) {
		TRACE_DEBUG(FULL, "Received an invalid msg or avp. EINVAL");
		return EINVAL;
	}

	/* Only accept objects with no linked children */
	if (!IS_LIST_EMPTY( &obj->children )) {
		TRACE_DEBUG(FULL, "Received an AVP or command with children. EINVAL");
		return EINVAL;
	}
	
	/* Unlink this object if needed */
	item_unlink( &obj->chaining );
	
	/* Free the octetstring if needed */
	if ((obj->type == _MSG_AVP) && (_A(obj)->avp_mustfreeos == 1)) {
		free(_A(obj)->avp_storage.os.data);
	}
	
	/* free the object */
	free(obj);
	
	return 0;
}

/***************************************************************************************************************/

/* Insert an item in a list, ordered by positions. */
static void insert_in_list_pos(_msg_list_t * sentinel, _msg_list_t * item)
{
	_msg_list_t *li;
	
	/* In case the item has no "position", assign the optional placement. This should never happen actually. */
	if (!item->pos)
		item->pos = RULE_OPTIONAL;
	
	/* Do some sanity checkings */
	assert(sentinel->head == sentinel); /* the sentinel is really a sentinel */
	assert(IS_LIST_EMPTY(item)); /* the item is not linked already */
	
	/* Find the place where the item should be inserted. */
	for (li = sentinel; (li->next != sentinel) && (li->next->pos <= item->pos); li = li->next);
	
	/* Now insert it here */
	item->head = sentinel;
	item->next = li->next;
	item->prev = li;
	li->next = item;
	item->next->prev = item;
}

/* Insert an item in a list, relatively to another item */
#define INSERT_after( __ref, __new ) {	\
	(__new)->next = (__ref)->next;	\
	(__new)->prev = (__ref);	\
	(__ref)->next->prev = (__new);	\
	(__ref)->next = (__new);	\
	(__new)->head = (__ref)->head;	\
}
#define INSERT_before( __ref, __new ) {	\
	(__new)->prev = (__ref)->prev;	\
	(__new)->next = (__ref);	\
	(__ref)->prev->next = (__new);	\
	(__ref)->prev = (__new);	\
	(__new)->head = (__ref)->head;	\
}
	
/* Unlink an item from a list */
static void item_unlink(_msg_list_t * item)
{
	/* Check first that this item is linked and not a sentinel */
	if (item->head == item)
		return;
	
	/* remove from the list */
	item->next->prev = item->prev;
	item->prev->next = item->next;
	
	/* sanitize the local values */
	item->next = item;
	item->prev = item;
	item->head = item;
}
	

/* Destroy an object and all its children */
static void destroy_tree(_msg_avp_chain_t * obj)
{
	_msg_list_t *sub;
	
	/* Destroy any subtree */
	while ( (sub = obj->children.next) != &obj->children)
		destroy_tree(sub->top);
	
	/* Then unlink and destroy the object */
	(void)destroy_obj(obj);
}


/***************************************************************************************************************/

/* We use this structure to pass arguments to create_child_avp when called in dict_iterate_rules */
typedef struct {
	_msg_avp_chain_t * parent;
	int flags;
} cb_data_t;

/* For each rule of a parent object, this function is called. This is part of the process of creations from templates */
static int create_child_avp(void *data, dict_rule_data_t *rule)
{
	int ret = 0, i;
	cb_data_t *_data = (cb_data_t *)data;

	TRACE_DEBUG(FULL, "enter (%p, %p)", _data, rule);
	
	/* Create as many children as the template says */
	for (i=0; i<rule->rule_template; i++) {
		msg_avp_t * child = NULL;
	
		/* Create the child AVP -- eventually a tree of AVPs */
		ret = msg_avp_new(rule->rule_avp, _data->flags, &child);
		if (ret != 0) {
			TRACE_DEBUG(FULL, "Creation of child AVP failed: %d (%s)", ret, strerror(ret));
			return ret;
		}

		/* If no error occurred, link this child into the parent's hierarchy */
		_C(child)->chaining.pos = rule->rule_position;
		insert_in_list_pos(&(_data->parent->children), &(_C(child)->chaining));
	}
		
	/* We're done with this rule entry */
	return 0;
}

/***************************************************************************************************************/
/* Debug functions: dumping */

/* indentation inside an object */
#define INOBJHDR 	"%*s   "
#define INOBJHDRVAL 	indent<0 ? 1 : indent, indent<0 ? "-" : "|"

/* Dump a msg_t object */
static void obj_dump_msg ( _msg_t * msg, int indent )
{
	int ret = 0;
	
	log_debug("%*sMSG: %p\n", INOBJHDRVAL, msg);
	if (!CHECK_MSG(msg)) {
		log_debug(INOBJHDR "INVALID!\n", INOBJHDRVAL);
		return;
	}
	if (!msg->msg_model) {
		log_debug(INOBJHDR "(no model)\n", INOBJHDRVAL);
	} else {
		dict_object_type_t dicttype;
		dict_cmd_data_t dictdata;
		ret = dict_gettype(msg->msg_model, &dicttype);
		if (ret || (dicttype != DICT_COMMAND)) {
			log_debug(INOBJHDR "(invalid model: %d %d)\n", INOBJHDRVAL, ret, dicttype);
			goto public;
		}
		ret = dict_getval(msg->msg_model, &dictdata);
		if (ret != 0) {
			log_debug(INOBJHDR "(error getting model data: %s)\n", INOBJHDRVAL, strerror(ret));
			goto public;
		}
		log_debug(INOBJHDR "model : v/m:" DUMP_CMDFL_str "/" DUMP_CMDFL_str ", %u \"%s\"\n", INOBJHDRVAL, 
			DUMP_CMDFL_val(dictdata.cmd_flag_val), DUMP_CMDFL_val(dictdata.cmd_flag_mask), dictdata.cmd_code, dictdata.cmd_name);
	}
public:	
	log_debug(INOBJHDR "public: V:%d L:%d fl:" DUMP_CMDFL_str " CC:%u A:%d hi:%d ei:%d\n", INOBJHDRVAL, 
		msg->msg_public.msg_version,
		msg->msg_public.msg_length,
		DUMP_CMDFL_val(msg->msg_public.msg_flags),
		msg->msg_public.msg_code,
		msg->msg_public.msg_appl,
		msg->msg_public.msg_hbhid,
		msg->msg_public.msg_eteid
		);
	log_debug(INOBJHDR "intern: src:%p rt:%d\n", INOBJHDRVAL, msg->msg_source, msg->msg_routable);
}
#define DUMP_VALUE(_format, _parms...)   log_debug(INOBJHDR "value : t:'%s' v:'" _format "'\n", INOBJHDRVAL, typename, ## _parms);
#define ASCII(_c) ( ((_c < 32) || (_c > 127)) ? ( _c ? '?' : ' ' ) : _c )
/* Dump an AVP value that is not a constant */
static void dump_basic_type(avp_value_t * value, dict_base_type_t type, char * typename, int indent)
{
	switch (type) {
		case AVP_TYPE_GROUPED:
			DUMP_VALUE("%s", "error: grouped AVP with a value!");
			break;
		case AVP_TYPE_OCTETSTRING:
			{
				/* Dump only up to 16 bytes of the buffer */
				unsigned char buf[16];
				memset(buf, 0, sizeof(buf));
				memcpy(buf, value->os.data, value->os.len < sizeof(buf)-1 ? value->os.len : sizeof(buf)-1 );
				DUMP_VALUE("l:%d, v:%02.2X %02.2X %02.2X %02.2X  %02.2X %02.2X %02.2X %02.2X  "
						   "%02.2X %02.2X %02.2X %02.2X  %02.2X %02.2X %02.2X %02.2X ... (%c%c%c%c %c%c%c%c %c%c%c%c %c%c%c%c ...)", 
						value->os.len,
						buf[0], buf[1], buf[2], buf[3], buf[4], buf[5], buf[6], buf[7], 
						buf[8], buf[9], buf[10], buf[11], buf[12], buf[13], buf[14], buf[15], 
						ASCII(buf[0]), ASCII(buf[1]), ASCII(buf[2]), ASCII(buf[3]), 
						ASCII(buf[4]), ASCII(buf[5]), ASCII(buf[6]), ASCII(buf[7]), 
						ASCII(buf[8]), ASCII(buf[9]), ASCII(buf[10]), ASCII(buf[11]), 
						ASCII(buf[12]), ASCII(buf[13]), ASCII(buf[14]), ASCII(buf[15])
						);
			}
			break;
		
		case AVP_TYPE_INTEGER32:
			DUMP_VALUE("%i",value->i32);
			break;

		case AVP_TYPE_INTEGER64:
			DUMP_VALUE("%li",value->i64);
			break;

		case AVP_TYPE_UNSIGNED32:
			DUMP_VALUE("%u",value->u32);
			break;

		case AVP_TYPE_UNSIGNED64:
			DUMP_VALUE("%lu",value->u64);
			break;

		case AVP_TYPE_FLOAT32:
			DUMP_VALUE("%g",(double)value->f32);
			break;

		case AVP_TYPE_FLOAT64:
			DUMP_VALUE("%g",value->f64);
			break;
		
		default:
			DUMP_VALUE("%s %d", "error: invalid type :", type);
	}
}
/* Dump an AVP value that is a constant */
#define DUMP_CONST(_format, _parms...)   log_debug(INOBJHDR "value : t:'%s' v: %s ( " _format " )\n", INOBJHDRVAL, typename, value->enum_name, ## _parms);
static void dump_constant_type(dict_type_enum_data_t * value, dict_base_type_t type, char * typename, int indent)
{
	switch (type) {
		case AVP_TYPE_GROUPED:
			DUMP_CONST("%s", "error: grouped AVP with a constant value!");
			break;
		case AVP_TYPE_OCTETSTRING:
			DUMP_CONST("%s", "value skipped");
			break;
		
		case AVP_TYPE_INTEGER32:
			DUMP_CONST("%i",value->enum_value.i32);
			break;

		case AVP_TYPE_INTEGER64:
			DUMP_CONST("%li",value->enum_value.i64);
			break;

		case AVP_TYPE_UNSIGNED32:
			DUMP_CONST("%u",value->enum_value.u32);
			break;

		case AVP_TYPE_UNSIGNED64:
			DUMP_CONST("%lu",value->enum_value.u64);
			break;

		case AVP_TYPE_FLOAT32:
			DUMP_CONST("%g",(double)value->enum_value.f32);
			break;

		case AVP_TYPE_FLOAT64:
			DUMP_CONST("%g",value->enum_value.f64);
			break;
		
		default:
			DUMP_CONST("%s %d", "error: invalid type :", type);
	}
}
/* Dump a msg_avp_t object */
static void obj_dump_avp ( _msg_avp_t * avp, int indent )
{
	int ret = 0;
	dict_base_type_t type = -1;
	
	log_debug("%*sAVP: %p\n", INOBJHDRVAL, avp);
	if (!CHECK_AVP(avp)) {
		log_debug(INOBJHDR "INVALID!\n", INOBJHDRVAL);
		return;
	}
	if (!avp->avp_model) {
		log_debug(INOBJHDR "(no model)\n", INOBJHDRVAL);
	} else {
		dict_object_type_t dicttype;
		dict_avp_data_t dictdata;
		ret = dict_gettype(avp->avp_model, &dicttype);
		if (ret || (dicttype != DICT_AVP)) {
			log_debug(INOBJHDR "(invalid model: %d %d)\n", INOBJHDRVAL, ret, dicttype);
			goto public;
		}
		ret = dict_getval(avp->avp_model, &dictdata);
		if (ret != 0) {
			log_debug(INOBJHDR "(error getting model data: %s)\n", INOBJHDRVAL, strerror(ret));
			goto public;
		}
		log_debug(INOBJHDR "model : v/m:" DUMP_AVPFL_str "/" DUMP_AVPFL_str ", %12s, %u \"%s\"\n", INOBJHDRVAL, 
			DUMP_AVPFL_val(dictdata.avp_flag_val), 
			DUMP_AVPFL_val(dictdata.avp_flag_mask), 
			type_base_name[dictdata.avp_basetype], 
			dictdata.avp_code, 
			dictdata.avp_name );
		type = dictdata.avp_basetype;
	}
public:	
	log_debug(INOBJHDR "public: C:%d fl:" DUMP_AVPFL_str " L:%d V:%d  data:@%p\n", INOBJHDRVAL, 
		avp->avp_public.avp_code,
		DUMP_AVPFL_val(avp->avp_public.avp_flags),
		avp->avp_public.avp_len,
		avp->avp_public.avp_vendor,
		avp->avp_public.avp_data
		);
	/* Dump the value if set */
	if (avp->avp_public.avp_data) {
		if (!avp->avp_model) {
			log_debug(INOBJHDR "(data set but no model: ERROR)\n", INOBJHDRVAL);
		} else {
			/* Try and find a constant name for this value */
			dict_object_t * avp_type = NULL;
			dict_object_t * avp_constant = NULL;
			dict_type_data_t type_data;
			dict_type_enum_request_t  request;
			ret = dict_search(DICT_TYPE, TYPE_OF_AVP, avp->avp_model, &avp_type);
			if ((ret != 0) || (avp_type == NULL)) {
				dump_basic_type(avp->avp_public.avp_data, type, type_base_name[type], indent);
				goto end;
			}
			ret = dict_getval(avp->avp_model, &type_data);
			if (ret != 0) {
				dump_basic_type(avp->avp_public.avp_data, type, "(error getting type data)", indent);
				goto end;
			}
			if (type_data.type_base != type) {
				dump_basic_type(avp->avp_public.avp_data, type, "(mismatching type information!)", indent);
				goto end;
			}
			/* Create a query for a constant */
			memset(&request, 0, sizeof(request));
			request.type_obj = avp_type;
			memcpy(&request.search.enum_value, avp->avp_public.avp_data, sizeof(avp_value_t));
			ret = dict_search(DICT_TYPE_ENUM, ENUM_BY_STRUCT, &request, &avp_constant);
			if ((ret != 0) || (avp_constant == NULL)) {
				dump_basic_type(avp->avp_public.avp_data, type, type_data.type_name, indent);
				goto end;
			}
			/* get the constant's information; we re-use request.search field */
			ret = dict_getval(avp_constant, &request.search);
			if (ret != 0) {
				dump_basic_type(avp->avp_public.avp_data, type, "(error getting constant data)", indent);
				goto end;
			}
			dump_constant_type(&request.search, type, type_data.type_name, indent);
		}
	}
end:	
	log_debug(INOBJHDR "intern: src:%p mf:%d\n", INOBJHDRVAL, avp->avp_source, avp->avp_mustfreeos);
}
/* Dump a single object content */
static void msg_dump_intern ( void * obj, int indent )
{
	/* Check the object */
	if (!VALIDATE_OBJ(obj)) {
		log_debug( ">>> invalid object (%p)!.\n", obj);
		return;
	}
	/* Dump the object */
	switch (_C(obj)->type) {
		case _MSG_AVP:
			obj_dump_avp ( _A(obj), indent );
			break;
		
		case _MSG_MSG:
			obj_dump_msg ( _M(obj), indent );
			break;
		
		default:
			assert(0);
	}
}
/***************************************************************************************************************/

/* Functions exported to other files in the daemon. More information in message.h */

/* Dump a message content -- for debug mostly */
void msg_dump_walk ( void * obj )
{
	void * ref = obj;
	void * next = NULL;
	int indent = 1;
	int ret = 0;
	
	log_debug("------ Dumping object %p (w)-------\n", obj);
	do {
		msg_dump_intern ( ref, indent );
		
		/* Now find the next object */
		ret = msg_browse ( ref, MSG_BRW_WALK, &next, &indent );
		if (ret != 0) {
			break;
		}
		/* dump next object */
	} while (ref = next);
	
out:
	log_debug("------ /end of object %p -------\n", obj);
}

/* Dump a single object content -- for debug mostly */
void msg_dump_one ( void * obj )
{
	log_debug("------ Dumping object %p (s)-------\n", obj);
	msg_dump_intern ( obj, 1 );
	log_debug("------ /end of object %p -------\n", obj);
}

/* Initialize the module */
int msg_init ( void )
{
	/* There are no globals to initialize in this module... */
	TRACE_DEBUG (INFO, "Message module initialized");
	return 0;
}

/* End of the module */
int msg_fini ( void )
{
	/* Nothing to do here */
	TRACE_DEBUG (FULL, "Unloading message module...");
	return 0;
}



/***************************************************************************************************************/

/* The API functions. See more information on these functions in message-api.h file. */

/* Create a new message instance */
int msg_new ( dict_object_t * model, int flags, msg_t ** msg )
{
	int ret=0;
	_msg_t *new = NULL;
	dict_object_type_t dicttype;
	dict_cmd_data_t    dictdata;
	
	TRACE_DEBUG (FULL, "Creating a message instance (%p, %x, %p)", model, flags, msg );
	
	/* Check the parameters */
	if ((!model) || (!msg) || (!CHECK_MSGFL(flags))) {
		TRACE_DEBUG (FULL, "EINVAL." );
		return EINVAL;
	}
	
	ret = dict_gettype(model, &dicttype);
	if (ret || (dicttype != DICT_COMMAND)) {
		TRACE_DEBUG (FULL, "(%d, %d) -> EINVAL.", ret, dicttype );
		return EINVAL;
	}
	
	ret = dict_getval(model, &dictdata);
	if (ret != 0) {
		TRACE_DEBUG (FULL, "Error occurred in dict_getval: %d (%s).", ret, strerror(ret) );
		return ret;
	}
	
	/* Create a new object */
	new = (_msg_t *) malloc (sizeof(_msg_t));
	if (!new) {
		log_error("Unable to allocate enough memory: %s\n", strerror(errno));
		TRACE_DEBUG (FULL, "malloc failed.");
		return ENOMEM;
	}
	
	/* Initialize the fields */
	init_msg(new);
	new->msg_model = model;
	
	new->msg_public.msg_version	= MSG_VERSION;
	new->msg_public.msg_length	= GETMSGHDRSZ(); /* This will be updated later */
	new->msg_public.msg_flags	= dictdata.cmd_flag_val;
	new->msg_public.msg_code	= dictdata.cmd_code;
	
	/* Create the children from template if needed */
	if (flags & MSGFL_FROM_TEMPLATE) {
		cb_data_t data = { _C(new), flags};
		ret = dict_iterate_rules ( model, &data, create_child_avp );
		if (ret != 0) {
			TRACE_DEBUG (FULL, "dict_iterate_rules failed (%s).", strerror(ret) );
			destroy_tree(_C(new));
			return ret;
		}
	}
	
	/* The new object is ready, return */
	*msg = (msg_t *)new;
	return 0;
}

/* Create a new AVP instance */
int msg_avp_new ( dict_object_t * model, int flags, msg_avp_t ** avp )
{
	int ret=0;
	_msg_avp_t *new = NULL;
	dict_object_type_t dicttype;
	dict_avp_data_t    dictavpdata;
	
	TRACE_DEBUG (FULL, "Creating an AVP instance (%p, %x, %p)", model, flags, avp );
	
	/* Check the parameters */
	if ((!model) || (!avp) || (!CHECK_MSGFL(flags))) {
		TRACE_DEBUG (FULL, "EINVAL." );
		return EINVAL;
	}
	
	ret = dict_gettype(model, &dicttype);
	if (ret || (dicttype != DICT_AVP)) {
		TRACE_DEBUG (FULL, "(%d, %d) -> EINVAL.", ret, dicttype );
		return EINVAL;
	}
	
	ret = dict_getval(model, &dictavpdata);
	if (ret != 0) {
		TRACE_DEBUG (FULL, "Error occurred in dict_getval: %d (%s).", ret, strerror(ret) );
		return ret;
	}
	
	/* Create a new object */
	new = (_msg_avp_t *) malloc (sizeof(_msg_avp_t));
	if (!new) {
		log_error("Unable to allocate enough memory: %s\n", strerror(errno));
		TRACE_DEBUG (FULL, "malloc failed.");
		return ENOMEM;
	}
	
	/* Initialize the fields */
	init_avp(new);
	new->avp_model = model;
	
	new->avp_public.avp_code    = dictavpdata.avp_code;
	new->avp_public.avp_flags   = dictavpdata.avp_flag_val;
	new->avp_public.avp_len = GETINITIALSIZE(dictavpdata.avp_basetype, dictavpdata.avp_flag_val & AVP_FLAG_VENDOR);
	new->avp_public.avp_vendor  = dictavpdata.avp_vendor;
	
	/* Create the children from template if needed */
	if ((flags & MSGFL_FROM_TEMPLATE) && (dictavpdata.avp_basetype == AVP_TYPE_GROUPED)) {
		cb_data_t data = { _C(new), flags};
		ret = dict_iterate_rules ( model, &data, create_child_avp );
		if (ret != 0) {
			TRACE_DEBUG (FULL, "dict_iterate_rules failed (%s).", strerror(ret) );
			destroy_tree(_C(new));
			return ret;
		}
	}
	
	/* The new object is ready, return */
	*avp = (msg_avp_t*)new;
	return 0;
}

/* Add an AVP into a tree */
int msg_avp_add ( void * reference, msg_dir_t dir, msg_avp_t *avp)
{
	/* Check the parameters */
	if (!VALIDATE_OBJ(reference)) {
		TRACE_DEBUG(FULL, "Received an invalid msg or avp. EINVAL");
		return EINVAL;
	}
	if (!CHECK_AVP(avp)) {
		TRACE_DEBUG(FULL, "Received an invalid avp. EINVAL");
		return EINVAL;
	}
	/* Check the avp to add is not chained */
	if (!IS_LIST_EMPTY(&_C(avp)->chaining)) {
		TRACE_DEBUG(FULL, "Received an avp already chained. EINVAL");
		return EINVAL;
	}
	
	/* Now insert */
	switch (dir) {
		case MSG_BRW_NEXT:
			/* Check the reference is an AVP -- we do not chain AVPs at same level as msgs. */
			if (_C(reference)->type != _MSG_AVP) {
				TRACE_DEBUG(FULL, "Received an invalid reference type. EINVAL");
				return EINVAL;
			}
			/* Insert the new avp after the reference */
			INSERT_after( &_C(reference)->chaining, &_C(avp)->chaining );
			break;

		case MSG_BRW_PREV:
			/* Check the reference is an AVP */
			if (_C(reference)->type != _MSG_AVP) {
				TRACE_DEBUG(FULL, "Received an invalid reference type. EINVAL");
				return EINVAL;
			}
			/* Insert the new avp before the reference */
			INSERT_before( &_C(reference)->chaining, &_C(avp)->chaining );
			break;

		case MSG_BRW_FIRST_CHILD:
			/* Insert the new avp after the children sentinel */
			INSERT_after( &_C(reference)->children, &_C(avp)->chaining );
			break;

		case MSG_BRW_LAST_CHILD:
			/* Insert the new avp before the children sentinel */
			INSERT_before( &_C(reference)->children, &_C(avp)->chaining );
			break;

		default:
			/* Other directions are invalid */
			TRACE_DEBUG(FULL, "Received an invalid direction. EINVAL");
			return EINVAL;
	}
			
	return 0;
}

/* Free an object and its tree */
int msg_free ( void * object, int subtree )
{
	if (!subtree)
		return destroy_obj ( _C(object) );
	
	destroy_tree(_C(object));
	return 0;
}

/* Allow exploring a message */
int msg_browse ( void * reference, msg_dir_t dir, void ** found, int * depth )
{
	int ret = 0;
	_msg_avp_chain_t *_found = NULL;
	int diff = 0;
	_msg_list_t      *li = NULL;
	
	/* Initialize the "found" result if any */
	if (found)
		*found = NULL;
	
	/* Check the parameters */
	if (!VALIDATE_OBJ(reference)) {
		TRACE_DEBUG(FULL, "Received an invalid reference object. EINVAL");
		return EINVAL;
	}
	
	TRACE_DEBUG(FULL, "chaining(%p): nxt:%p prv:%p hea:%p top:%p pos:%d", 
			&_C(reference)->chaining,
			_C(reference)->chaining.next,
			_C(reference)->chaining.prev,
			_C(reference)->chaining.head,
			_C(reference)->chaining.top,
			_C(reference)->chaining.pos);
	TRACE_DEBUG(FULL, "children(%p): nxt:%p prv:%p hea:%p top:%p pos:%d", 
			&_C(reference)->children,
			_C(reference)->children.next,
			_C(reference)->children.prev,
			_C(reference)->children.head,
			_C(reference)->children.top,
			_C(reference)->children.pos);

	/* Now search */
	switch (dir) {
		case MSG_BRW_NEXT:
			/* Check the reference is an AVP */
			if (_C(reference)->type != _MSG_AVP) {
				TRACE_DEBUG(FULL, "Received an invalid reference type. EINVAL");
				return EINVAL;
			}

			li = &_C(reference)->chaining;
			
			/* Check if the next element is not the sentinel ( ==> the parent) */
			if (li->next != li->head)
				_found = li->next->top;
			else
				ret = ENOENT; /* there is no next element */
			break;

		case MSG_BRW_PREV:
			/* Check the reference is an AVP */
			if (_C(reference)->type != _MSG_AVP) {
				TRACE_DEBUG(FULL, "Received an invalid reference type. EINVAL");
				return EINVAL;
			}

			li = &_C(reference)->chaining;
			
			/* Check if the prev element is not the sentinel ( ==> the parent) */
			if (li->prev != li->head)
				_found = li->prev->top;
			else
				ret = ENOENT;
			break;

		case MSG_BRW_FIRST_CHILD:
			li = &_C(reference)->children;
			if (IS_LIST_EMPTY(li)) {
				/* No children */
				ret = ENOENT;
			} else {
				_found = li->next->top;
				diff = 1;
			}
			break;

		case MSG_BRW_LAST_CHILD:
			li = &_C(reference)->children;
			if (IS_LIST_EMPTY(li)) {
				/* No children */
				ret = ENOENT;
			} else {
				_found = li->prev->top;
				diff = 1;
			}
			break;

		case MSG_BRW_PARENT:
			/* If the object is not chained, it has no parent */
			li = &_C(reference)->chaining;
			if (IS_LIST_EMPTY(li)) {
				/* No parent */
				ret = ENOENT;
			} else {
				/* The sentinel is the parent's children list */
				_found = li->head->top;
				diff = -1;
			}
			break;

		case MSG_BRW_WALK:
			/* First, try to find a child */
			li = &_C(reference)->children;
			if ( ! IS_LIST_EMPTY(li) ) {
				_found = li->next->top;
				diff = 1;
				break;
			}
			
			/* Then try to find a "next" at this level or one of the parent's */
			li = &_C(reference)->chaining;
			do {
				/* If this element has a "next" element, return it */
				if (li->next != li->head) {
					_found = li->next->top;
					break;
				}
				/* otherwise, check if we have a parent */
				if (IS_LIST_EMPTY(li)) {
					/* no parent */
					ret = ENOENT;
					break;
				}
				/* Go to the parent's chaining information and loop */
				diff -= 1;
				li = &li->head->top->chaining;
			} while (1); 
			break;
			
		default:
			/* Other directions are invalid */
			TRACE_DEBUG(FULL, "Received an invalid direction. EINVAL");
			return EINVAL;
	}
	
	/* Save the found object, if any */
	if (found && _found)
		*found = (void *)_found;
	/* Modify the depth according to the wlak direction */
	if (depth && diff)
		(*depth) += diff;
	
	/* Return the error code */
	return ret;
}

/* Retrieve the model of an object */
int msg_model ( void * reference, dict_object_t ** model )
{
	/* Check the parameters */
	if ((!model) || (!VALIDATE_OBJ(reference))) {
		TRACE_DEBUG(FULL, "Received an invalid reference object. EINVAL");
		return EINVAL;
	}
	
	/* Dump the object */
	switch (_C(reference)->type) {
		case _MSG_AVP:
			*model = _A(reference)->avp_model;
			break;
		
		case _MSG_MSG:
			*model = _M(reference)->msg_model;
			break;
		
		default:
			assert(0);
	}
	
	return 0;
}

/* Set the value of an AVP */
int msg_avp_setvalue ( msg_avp_t *avp, avp_value_t *value )
{
	dict_base_type_t type = -1;
	
	if ((!CHECK_AVP(avp)) || (!_A(avp)->avp_model)) {
		TRACE_DEBUG(FULL, "Received an invalid reference object. EINVAL");
		return EINVAL;
	}
	
	{
		dict_object_type_t dicttype = -1;
		dict_avp_data_t dictdata;
		int ret = 0;
		ret = dict_gettype(_A(avp)->avp_model, &dicttype);
		if (ret || (dicttype != DICT_AVP)) {
			TRACE_DEBUG(FULL, "Received an invalid reference object. EINVAL");
			return EINVAL;
		}
		ret = dict_getval(_A(avp)->avp_model, &dictdata);
		if (ret != 0) {
			TRACE_DEBUG(FULL, "Error getting avp model data");
			return ret;
		}
		type = dictdata.avp_basetype;
		if (type == AVP_TYPE_GROUPED) {
			TRACE_DEBUG(FULL, "Cannot set a value to a grouped AVP. EINVAL");
			return EINVAL;
		}
	}
	
	/* First, clean any previous value */
	if (_A(avp)->avp_mustfreeos != 0) {
		free(_A(avp)->avp_storage.os.data);
		_A(avp)->avp_mustfreeos = 0;
	}
	
	memset(&_A(avp)->avp_storage, 0, sizeof(avp_value_t));
	
	if (!value) {
		_A(avp)->avp_public.avp_data = NULL;
		return 0;
	}
	
	/* Now we have to set the value */
	memcpy(&_A(avp)->avp_storage, value, sizeof(avp_value_t));
	
	/* Copy an octetstring if needed. */
	if (type == AVP_TYPE_OCTETSTRING) {
		_A(avp)->avp_storage.os.data = malloc(value->os.len);
		if (!_A(avp)->avp_storage.os.data) {
			log_error("Unable to allocate enough memory: %s\n", strerror(errno));
			TRACE_DEBUG (FULL, "malloc failed.");
			return ENOMEM;
		}
		_A(avp)->avp_mustfreeos = 1;
		memcpy(_A(avp)->avp_storage.os.data, value->os.data, value->os.len);
	}
	
	/* Set the data pointer of the public part */
	_A(avp)->avp_public.avp_data = &_A(avp)->avp_storage;
	
	return 0;		
}

/* Retrieve the address of the msg_public field of a message */
int msg_data ( msg_t *msg, msg_data_t **pdata )
{
	if ((!CHECK_MSG(msg)) || (!pdata)) {
		TRACE_DEBUG(FULL, "Received an invalid reference object. EINVAL");
		return EINVAL;
	}
	*pdata = & _M(msg)->msg_public;
	return 0;
}

/* Retrieve the address of the avp_public field of an avp */
int msg_avp_data ( msg_avp_t *avp, msg_avp_data_t **pdata )
{
	if ((!CHECK_AVP(avp)) || (!pdata)) {
		TRACE_DEBUG(FULL, "Received an invalid reference object. EINVAL");
		return EINVAL;
	}
	*pdata = & _A(avp)->avp_public;
	return 0;
}

/* This macro will pad a size to the next multiple of 4. */
#define PAD4(_x) ((_x) + ( (4 - (_x)) & 3 ) )

/* Compute the lengh of an object and its subtree. */
int msg_update_length ( void * object )
{
	int ret = 0;
	size_t sz = 0;
	dict_object_t * model;
	
	union {
		dict_cmd_data_t cmddata;
		dict_avp_data_t avpdata;
	} dictdata;
	
	/* Get the model of the object. This also validates the object */
	ret = msg_model ( object, &model );
	/* Check the object is valid */
	if ((ret != 0) || (model == NULL)) {
		TRACE_DEBUG(FULL, "Received an invalid object, or object without model. EINVAL");
		return EINVAL;
	}
	
	/* Get the information of the model */
	ret = dict_getval(model, &dictdata);
	if (ret != 0) {
		TRACE_DEBUG(FULL, "Error in dict_getdata: %s", strerror(ret));
		return ret;
	}
	
	/* Deal with easy cases: AVPs without children */
	if ((_C(object)->type == _MSG_AVP) && (dictdata.avpdata.avp_basetype != AVP_TYPE_GROUPED)) {
		/* Sanity check */
		assert(IS_LIST_EMPTY(&_C(object)->children));
		
		/* Now check that the data is set in the AVP */
		if (_A(object)->avp_public.avp_data == NULL) {
			TRACE_DEBUG(FULL, "The avp_data is not set.");
			return EINVAL;
		}
		
		sz = GETAVPHDRSZ( _A(object)->avp_public.avp_flags & AVP_FLAG_VENDOR );
		
		switch (dictdata.avpdata.avp_basetype) {
			case AVP_TYPE_OCTETSTRING:
				sz += _A(object)->avp_public.avp_data->os.len;
				break;
			
			case AVP_TYPE_INTEGER32:
			case AVP_TYPE_INTEGER64:
			case AVP_TYPE_UNSIGNED32:
			case AVP_TYPE_UNSIGNED64:
			case AVP_TYPE_FLOAT32:
			case AVP_TYPE_FLOAT64:
				sz += avp_value_sizes[dictdata.avpdata.avp_basetype];
				break;
			
			default:
				/* Something went wrong... */
				assert(0);
		}
	}
	else  /* message or grouped AVP */
	{
		_msg_list_t * ch = NULL;
		
		/* Compute the header size first */
		if (_C(object)->type == _MSG_AVP) 
			sz = GETAVPHDRSZ( _A(object)->avp_public.avp_flags & AVP_FLAG_VENDOR );
		else
			sz = GETMSGHDRSZ( );
		
		/* Recurse in all children and update the sz information */
		for (ch = _C(object)->children.next; ch != &_C(object)->children; ch = ch->next) {
			ret = msg_update_length ( ch->top );
			if (ret != 0) {
				TRACE_DEBUG(FULL, "msg_update_length failed in child %p", ch->next->top);
				return ret;
			}
			
			/* Pad any previous size to 32b boundary, and add the child length */
			sz += PAD4( sz ) + _A(ch->top)->avp_public.avp_len;
		}
	}
	
	/* If we arrive here, the "sz" variable contains the size to write in the object */
	if (_C(object)->type == _MSG_AVP) 
		_A(object)->avp_public.avp_len = sz;
	else
		_M(object)->msg_public.msg_length = sz;
	
	return 0;
}

/**********************************************************************************************************/
/* The following functions are used to parse messages from and to buffers. It is responsible for converting 
 * to and from network byte-order.
 */

int msg_parse_buffer ( char * buffer, msg_t ** msg )
{
	
	TRACE_DEBUG (FULL, "@@@ %s: not implemented yet.", __FUNCTION__ );
	return ENOTSUP;
}

int msg_bufferize ( msg_t * msg, char ** buffer, int * is_routable )
{
	TRACE_DEBUG (FULL, "@@@ %s: not implemented yet.", __FUNCTION__ );
	return ENOTSUP;
}



int msg_parse_dict ( msg_t * msg )
{
	TRACE_DEBUG (FULL, "@@@ %s: not implemented yet.", __FUNCTION__ );
	return ENOTSUP;
}




int msg_avp_getrawdata ( msg_avp_t *avp, char **data )
{
	TRACE_DEBUG (FULL, "@@@ %s: not implemented yet.", __FUNCTION__ );
	return ENOTSUP;
}


int msg_parse_rules ( void * object, dict_object_t ** rule, msg_avp_t **avp)
{
	TRACE_DEBUG (FULL, "@@@ %s: not implemented yet.", __FUNCTION__ );
	return ENOTSUP;
}



